﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;

namespace imlac
{
    
    public class Debugger
    {
        public Debugger(ImlacSystem system)
        {
            _system = system;
        }

        public SystemExecutionState Prompt()
        {
            SystemExecutionState nextState = SystemExecutionState.Debugging;

            PrintProcessorStatus();            

            Console.Write(">");

            String input = Console.ReadLine().ToLower().Trim();

            if (string.IsNullOrWhiteSpace(input))
            {                
                input = _lastInput;
                
                // echo the last input
                Console.WriteLine(_lastInput);
            }

            String[] tokens = input.Split(' ');

            if (tokens.Length > 0)
            {
                switch (tokens[0])
                {
                    case "?":
                        //PrintHelp();
                        break;

                    case "r":
                        _system.Reset();
                        Console.WriteLine("System reset.");
                        break;

                    case "m":
                        if (tokens.Length == 2)
                        {
                            MemoryOperation(Helpers.GetUshortForOctalString(tokens[1]));
                        }
                        break;                    

                    case "g":
                        if (tokens.Length == 2)
                        {
                            _system.Processor.PC = Helpers.GetUshortForOctalString(tokens[1]);
                        }
                        _system.Processor.State = ProcessorState.Running;
                        nextState = SystemExecutionState.Running;
                        break;

                    case "s":
                        if (tokens.Length == 1)
                        {
                            _system.Processor.State = ProcessorState.Running;
                            nextState = SystemExecutionState.SingleStep;
                        }
                        else if (tokens.Length == 2)
                        {
                            _system.Processor.PC = Helpers.GetUshortForOctalString(tokens[1]);
                            _system.Processor.State = ProcessorState.Running;
                            nextState = SystemExecutionState.SingleStep;
                        }
                        break;

                    case "f":
                        _system.Processor.State = ProcessorState.Running;
                        nextState = SystemExecutionState.SingleFrame;
                        break;

                    case "fs":
                        _system.Processor.State = ProcessorState.Running;
                        nextState = SystemExecutionState.UntilDisplayStart;
                        break;

                    case "b":
                        if (tokens.Length == 2)
                        {
                            try
                            {
                                LoadMemory(Paths.BuildBootPath(tokens[1]), 0x20, 0x20);
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }
                        }                        
                        break;                        

                    case "sc":
                        if (tokens.Length == 2)
                        {
                            try
                            {
                                SaveMemory(tokens[1], 0, _system.Memory.Size);
                            }
                            catch(Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }
                        }
                        else if( tokens.Length == 4)
                        {                        
                            try
                            {
                                SaveMemory(tokens[1], Helpers.GetUshortForOctalString(tokens[2]), Helpers.GetUshortForOctalString(tokens[3]));
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }
                        
                        }
                        break;

                    case "lc":
                        if (tokens.Length == 2)
                        {
                            try
                            {
                                LoadMemory(tokens[1], 0, _system.Memory.Size);
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }
                        }
                        else if (tokens.Length == 4)
                        {
                            try
                            {
                                LoadMemory(tokens[1], Helpers.GetUshortForOctalString(tokens[2]), Helpers.GetUshortForOctalString(tokens[3]));
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }

                        }
                        break;

                    case "p":
                        if (tokens.Length == 2)
                        {
                            try
                            {
                                _system.PaperTapeReader.LoadTape(tokens[1]);
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }
                        }
                        break;

                    case "t":
                        if (tokens.Length == 2)
                        {
                            try
                            {
                                _system.TTY.LoadImage(tokens[1]);
                            }
                            catch (Exception e)
                            {
                                Console.WriteLine(e.Message);
                            }
                        }
                        break;

                    case "u":
                        Disassemble(tokens, DisassemblyMode.Processor);
                        break;

                    case "ud":
                        Disassemble(tokens, DisassemblyMode.DisplayAuto);
                        break;

                    case "udp":
                        Disassemble(tokens, DisassemblyMode.DisplayProcessor);
                        break;

                    case "udi":
                        Disassemble(tokens, DisassemblyMode.DisplayIncrement);
                        break;

                    case "d":
                        if (tokens.Length == 2)
                        {
                            _system.Processor.DS = Helpers.GetUshortForOctalString(tokens[1]);
                        }
                        else
                        {
                            Console.WriteLine(Helpers.ToOctal(_system.Processor.DS));
                        }
                        break;

                    case "l":
                        if (tokens.Length == 2)
                        {
                            ushort traceFlags = Helpers.GetUshortForOctalString(tokens[1]);
                            Trace.TraceLevel = (LogType)traceFlags;

                            Trace.TraceOn = traceFlags != 0;
                        }
                        else
                        {
                            Console.WriteLine(Helpers.ToOctal((ushort)Trace.TraceLevel));
                        }
                        break;

                    case "dm":
                        if (tokens.Length == 2)
                        {
                            DumpMemory(Helpers.GetUshortForOctalString(tokens[1]), 1);
                        }
                        else if (tokens.Length == 3)
                        {
                            DumpMemory(Helpers.GetUshortForOctalString(tokens[1]), Helpers.GetUshortForOctalString(tokens[2]));
                        }
                        break;

                    case "cd":
                        _system.Display.ClearDisplay();
                        break;

                    default:
                        Console.WriteLine("?");     // ken would approve.
                        break;
                }
            }

            _lastInput = input;

            return nextState;
        }

        private void PrintProcessorStatus()
        {
            Console.WriteLine("PC={0} AC={1} MB={2} - {3}",
                Helpers.ToOctal(_system.Processor.PC),
                Helpers.ToOctal(_system.Processor.AC),
                Helpers.ToOctal(_system.Memory.Fetch(_system.Processor.PC)),
                _system.Processor.State);

            Console.WriteLine("DPC={0} DT={1} DPCE={2} X={3} Y={4}\nMode={5} HalfWord={6} - {7}",
                Helpers.ToOctal(_system.DisplayProcessor.PC),
                Helpers.ToOctal(_system.DisplayProcessor.DT),
                Helpers.ToOctal(_system.DisplayProcessor.DPCEntry),
                _system.DisplayProcessor.X,
                _system.DisplayProcessor.Y,
                _system.DisplayProcessor.Mode,
                _system.DisplayProcessor.Half,
                _system.DisplayProcessor.State);
        }

        private void MemoryOperation(ushort address)
        {
            bool done = false;

            while (!done)
            {
                try
                {
                    Console.Write("{0}\\{1} ", Helpers.ToOctal(address), Helpers.ToOctal(_system.Memory.Fetch(address)));
                    string dataString = Console.ReadLine();

                    if (!string.IsNullOrWhiteSpace(dataString))
                    {
                        ushort value = Helpers.GetUshortForOctalString(dataString);
                        _system.Memory.Store(address, value);                        
                    }

                    address++;
                }
                catch
                {
                    done = true;
                }
            }
        }

        private void DumpMemory(ushort startAddress, int count)
        {
            while (count > 0)
            {
                Console.Write("{0}: ", Helpers.ToOctal(startAddress));
                for (ushort address = startAddress; address < startAddress + 4; address++)
                {
                    Console.Write("{0} ", Helpers.ToOctal(_system.Memory.Fetch(address)));
                }

                for (ushort address = startAddress; address < startAddress + 4; address++)
                {
                    ushort word = _system.Memory.Fetch(address);
                    Console.Write("{0}{1} ", GetPrintableChar((char)(word >> 8)), GetPrintableChar((char)(word & 0xff)));
                }

                Console.WriteLine();

                startAddress += 4;
                count -= 4;
            }
        }

        private char GetPrintableChar(char c)
        {
            if (char.IsLetterOrDigit(c) ||
                char.IsPunctuation(c) ||
                char.IsSymbol(c))
            {
                return c;
            }
            else
            {
                return '.';
            }
        }

        private void Disassemble(string[] tokens, DisassemblyMode mode)
        {
            ushort startAddress;
            ushort endAddress;
            if (tokens.Length == 1)
            {
                // disassemble just the current instruction
                startAddress = endAddress = _system.Processor.PC;
                
            }
            else if (tokens.Length == 2)
            {
                // disassemble the specified address
                startAddress = endAddress = Helpers.GetUshortForOctalString(tokens[1]);                               
            }
            else if (tokens.Length == 3)
            {
                // disassemble from specified start to specified end address
                startAddress = Helpers.GetUshortForOctalString(tokens[1]);
                endAddress = Helpers.GetUshortForOctalString(tokens[2]);   
            }
            else
            {
                // not a valid command
                return;
            }

            for (ushort address = startAddress; address <= endAddress; address++)
            {
                string disassembly = string.Empty;
                try
                {
                    switch (mode)
                    {
                        case DisassemblyMode.Processor:
                            disassembly = _system.Processor.Disassemble(address);
                            break;

                        case DisassemblyMode.DisplayAuto:
                            disassembly = _system.DisplayProcessor.Disassemble(address, DisplayProcessorMode.Indeterminate);
                            break;

                        case DisassemblyMode.DisplayProcessor:
                            disassembly = _system.DisplayProcessor.Disassemble(address, DisplayProcessorMode.Processor);
                            break;

                        case DisassemblyMode.DisplayIncrement:
                            disassembly = _system.DisplayProcessor.Disassemble(address, DisplayProcessorMode.Increment);
                            break;
                    }
                }
                catch
                {
                    // this can happen if the data is not a valid instruction
                }

                Console.WriteLine("{0}\\{1} {2}", Helpers.ToOctal(address), Helpers.ToOctal(_system.Memory.Fetch(address)), disassembly);
            }
        }

        private void SaveMemory(string path, ushort startAddress, ushort length)
        {
            if (startAddress > _system.Memory.Size)
            {
                throw new InvalidOperationException(String.Format("Start address must be less than the size of system memory ({0}).", Helpers.ToOctal(_system.Memory.Size)));
            }

            // clip end to top of memory
            ushort endAddress = (ushort)Math.Min(_system.Memory.Size, startAddress + length);

            FileStream fs = File.OpenWrite(path);

            for (ushort i = startAddress; i < endAddress; i++)
            {
                ushort value = _system.Memory.Fetch(i);

                fs.WriteByte((byte)(value >> 8));
                fs.WriteByte((byte)(value & 0xff));
            }

            fs.Close();
        }

        private void LoadMemory(string path, ushort startAddress, ushort length)
        {
            if (startAddress > _system.Memory.Size)
            {
                throw new InvalidOperationException(String.Format("Start address must be less than the size of system memory ({0}).", Helpers.ToOctal(_system.Memory.Size)));
            }
            
            // clip end to top of memory
            ushort endAddress = (ushort)Math.Min(_system.Memory.Size, startAddress + length);

            FileStream fs = File.OpenRead(path);

            for (ushort i = startAddress; i < endAddress; i++)
            {
                ushort value = (ushort)((fs.ReadByte() << 8) | (fs.ReadByte()));
                _system.Memory.Store(i, value);
            }

            fs.Close();
        }

        private enum DisassemblyMode
        {
            Processor,
            DisplayProcessor,
            DisplayIncrement,
            DisplayAuto
        }

        private ImlacSystem _system;
        private string _lastInput;
    }
}
