﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;

using imlac.IO;
using SdlDotNet.Core;
using SdlDotNet.Graphics;
using SdlDotNet.Windows;
using SdlDotNet.Input;

namespace imlac
{
    /// <summary>
    /// This is used to filter keyboard messages from the App's message loop.
    /// This is necessary because the "Video" object that SDL provides appears
    /// to eat certain keystrokes and since I have no means to control it
    /// (and I'm not really all that interested in compiling my own version)
    /// this is the only alternative.
    /// 
    /// This class captures keystrokes for keys the emulator is interested in
    /// and fires an event containing the keycode.
    /// These keyboard messages are always passed on to the app.
    /// </summary>
    public class KeyboardFilter : IMessageFilter
    {
        public KeyboardFilter()
        {
            _keyModifiers = ImlacKeyModifiers.None;
            _keyLatched = false;
            _dataSwitches = 0xffff;
        }

        static KeyboardFilter()
        {
            BuildKeyMappings();
        }        

        public event EventHandler FullScreenToggle; 

        // TODO: should ensure consistency (threading)

        public ImlacKey Key
        {
            get { return _keyCode; }
        }

        public ImlacKeyModifiers Modifiers
        {
            get { return _keyModifiers; }
        }

        public bool KeyLatched
        {
            get { return _keyLatched; }
            set { _keyLatched = value; }
        }

        public ushort DataSwitches
        {
            get { return (ushort)_dataSwitches; }
        }

        public bool PreFilterMessage(ref Message m)
        {
            bool ret = false;

            switch (m.Msg)
            {
                case WM_SYSKEYDOWN:
                case WM_SOMETHINGDOWN:
                    //
                    // If this is a modifier key (Alt, Shift, Ctrl) then we track it separately
                    // (it is not tracked as a key)
                    //                    
                    switch ((VKeys)m.WParam.ToInt32())
                    {
                        case VKeys.Shift:
                            _keyModifiers |= ImlacKeyModifiers.Shift;
                            Console.WriteLine("shift");
                            break;

                        case VKeys.Ctrl:
                            _keyModifiers |= ImlacKeyModifiers.Ctrl;
                            Console.WriteLine("ctrl");
                            break;

                        case VKeys.Alt:
                            _keyModifiers |= ImlacKeyModifiers.Rept;
                            Console.WriteLine("alt");
                            break;

                        default:

                            UpdateDataSwitches((VKeys)m.WParam.ToInt32(), true /* key down */);

                            //Console.WriteLine("{0:x}", m.WParam.ToInt32());

                            if ((VKeys)m.WParam.ToInt32() == VKeys.Insert)
                            {
                                if (FullScreenToggle != null)
                                {
                                    FullScreenToggle(this, null);
                                }                                
                            }

                            if (!_keyLatched)
                            {
                                _keyCode = TranslateKeyCode((VKeys)m.WParam.ToInt32());

                                //Console.WriteLine("{0:x} {1},{2}", m.WParam.ToInt32(), _keyCode, _keyModifiers);

                                if (_keyCode != ImlacKey.Invalid)
                                {
                                    _keyLatched = true;
                                }
                            }
                            break;
                    }
                    break;

                case WM_SYSKEYUP:
                case WM_SOMETHINGUP:
                    //
                    // We only track keyboard modifiers and data switch toggles here.
                    //
                    switch ((VKeys)m.WParam.ToInt32())
                    {
                        case VKeys.Shift:
                            _keyModifiers &= (~ImlacKeyModifiers.Shift);
                            break;

                        case VKeys.Ctrl:
                            _keyModifiers &= (~ImlacKeyModifiers.Ctrl);
                            break;

                        case VKeys.Alt:
                            _keyModifiers &= (~ImlacKeyModifiers.Rept);
                            break;

                        default:
                            UpdateDataSwitches((VKeys)m.WParam.ToInt32(), false /* key up */);
                            break;
                    }
                    break;

                default:
                    //Console.WriteLine("msg {0:x} lp {1:x} wp {2:x}", m.Msg, m.LParam.ToInt32(), m.WParam.ToInt32());
                    break;
            }

            return ret;
        }

        private ImlacKey TranslateKeyCode(VKeys virtualKey)
        {
            if (_keyMappings.ContainsKey(virtualKey))
            {
                return _keyMappings[virtualKey];
            }
            else
            {
                return ImlacKey.Invalid;
            }
        }

        private void UpdateDataSwitches(VKeys virtualKey, bool keyDown)
        {
            //
            // If this is a key mapped to a front panel switch
            // we will toggle the bit in the DS register based on whether
            // the key is down or up
            //
            for (int i = 0; i < 16; i++)
            {
                if (virtualKey == _dataSwitchMappings[i])
                {
                    // toggle this bit
                    if (!keyDown)
                    {
                        // or it in
                        _dataSwitches |= (0x1 << (15 - i));
                    }
                    else
                    {
                        // mask it out
                        _dataSwitches &= ~(0x1 << (15 - i));
                    }

                    Console.WriteLine("{0:x}", _dataSwitches);
                    break;
                }
            }
        }

        private static void BuildKeyMappings()
        {
            _keyMappings = new Dictionary<VKeys, ImlacKey>();

            _keyMappings.Add(VKeys.End, ImlacKey.DataXmit);
            _keyMappings.Add(VKeys.DownArrow, ImlacKey.Down);
            _keyMappings.Add(VKeys.RightArrow, ImlacKey.Right);
            _keyMappings.Add(VKeys.UpArrow, ImlacKey.Up);
            _keyMappings.Add(VKeys.LeftArrow, ImlacKey.Left);
            _keyMappings.Add(VKeys.Tab, ImlacKey.Tab);
            _keyMappings.Add(VKeys.Return, ImlacKey.CR);
            _keyMappings.Add(VKeys.PageUp, ImlacKey.FF);
            _keyMappings.Add(VKeys.PageDown, ImlacKey.PageXmit);
            _keyMappings.Add(VKeys.Home, ImlacKey.Home);
            _keyMappings.Add(VKeys.Pause, ImlacKey.Brk);
            _keyMappings.Add(VKeys.Escape, ImlacKey.Esc);
            _keyMappings.Add(VKeys.Space, ImlacKey.Space);

            _keyMappings.Add(VKeys.Comma, ImlacKey.Comma);
            _keyMappings.Add(VKeys.Plus, ImlacKey.Minus);
            _keyMappings.Add(VKeys.Period, ImlacKey.Period);
            _keyMappings.Add(VKeys.QuestionMark, ImlacKey.Slash);
            _keyMappings.Add(VKeys.Zero, ImlacKey.K0);
            _keyMappings.Add(VKeys.One, ImlacKey.K1);
            _keyMappings.Add(VKeys.Two, ImlacKey.K2);
            _keyMappings.Add(VKeys.Three, ImlacKey.K3);
            _keyMappings.Add(VKeys.Four, ImlacKey.K4);
            _keyMappings.Add(VKeys.Five, ImlacKey.K5);
            _keyMappings.Add(VKeys.Six, ImlacKey.K6);
            _keyMappings.Add(VKeys.Seven, ImlacKey.K7);
            _keyMappings.Add(VKeys.Eight, ImlacKey.K8);
            _keyMappings.Add(VKeys.Nine, ImlacKey.K9);
            _keyMappings.Add(VKeys.Minus, ImlacKey.Colon);
            _keyMappings.Add(VKeys.Semicolon, ImlacKey.Semicolon);

            _keyMappings.Add(VKeys.Keypad0, ImlacKey.D0);
            _keyMappings.Add(VKeys.Keypad2, ImlacKey.D2);
            _keyMappings.Add(VKeys.Keypad4, ImlacKey.D4);
            _keyMappings.Add(VKeys.Keypad5, ImlacKey.D5);
            _keyMappings.Add(VKeys.Keypad6, ImlacKey.D6);
            _keyMappings.Add(VKeys.DoubleQuote, ImlacKey.Unlabeled);

            _keyMappings.Add(VKeys.A, ImlacKey.A);
            _keyMappings.Add(VKeys.B, ImlacKey.B);
            _keyMappings.Add(VKeys.C, ImlacKey.C);
            _keyMappings.Add(VKeys.D, ImlacKey.D);
            _keyMappings.Add(VKeys.E, ImlacKey.E);
            _keyMappings.Add(VKeys.F, ImlacKey.F);
            _keyMappings.Add(VKeys.G, ImlacKey.G);
            _keyMappings.Add(VKeys.H, ImlacKey.H);
            _keyMappings.Add(VKeys.I, ImlacKey.I);
            _keyMappings.Add(VKeys.J, ImlacKey.J);
            _keyMappings.Add(VKeys.K, ImlacKey.K);
            _keyMappings.Add(VKeys.L, ImlacKey.L);
            _keyMappings.Add(VKeys.M, ImlacKey.M);
            _keyMappings.Add(VKeys.N, ImlacKey.N);
            _keyMappings.Add(VKeys.O, ImlacKey.O);
            _keyMappings.Add(VKeys.P, ImlacKey.P);
            _keyMappings.Add(VKeys.Q, ImlacKey.Q);
            _keyMappings.Add(VKeys.R, ImlacKey.R);
            _keyMappings.Add(VKeys.S, ImlacKey.S);
            _keyMappings.Add(VKeys.T, ImlacKey.T);
            _keyMappings.Add(VKeys.U, ImlacKey.U);
            _keyMappings.Add(VKeys.V, ImlacKey.V);
            _keyMappings.Add(VKeys.W, ImlacKey.W);
            _keyMappings.Add(VKeys.X, ImlacKey.X);
            _keyMappings.Add(VKeys.Y, ImlacKey.Y);
            _keyMappings.Add(VKeys.Z, ImlacKey.Z);

            _keyMappings.Add(VKeys.Delete, ImlacKey.Del);
        }

        private static Dictionary<VKeys, ImlacKey> _keyMappings;

        private ImlacKey            _keyCode;
        private ImlacKeyModifiers   _keyModifiers;
        private bool                _keyLatched;

        // Data switch mappings (TODO: make configurable)
        // There are 16 switches mapped here, a value of 0 or 1 means
        // that no key is mapped, hardcode return value to 0 or 1
        // The first entry corresponds to bit 0 (MSB)
        // and the last entry corresponds to bit 15 (LSB)
        private VKeys[] _dataSwitchMappings = 
        {
            VKeys.F1,
            VKeys.F2,
            VKeys.F3,
            VKeys.F4,
            0,
            0,
            0,
            0,
            VKeys.F5,
            VKeys.F6,
            VKeys.F7,
            VKeys.F8,

            VKeys.F9,
            VKeys.F10,
            VKeys.F11,
            VKeys.F12,
        };

        private int _dataSwitches;

        private const int WM_SOMETHINGUP = 0x105;
        private const int WM_SOMETHINGDOWN = 0x104;
        private const int WM_KEYUP = 0x103;
        private const int WM_KEYDOWN = 0x102;
        private const int WM_SYSKEYUP = 0x101;
        private const int WM_SYSKEYDOWN = 0x100;

        private enum VKeys
        {
            // Keys that map to Imlac keyboard keys
            Shift = 0x10,
            Ctrl = 0x11,
            Alt = 0x12,

            End = 0x23,
            DownArrow = 0x28,
            RightArrow = 0x27,
            UpArrow = 0x26,
            LeftArrow = 0x25,
            Tab = 0x9,
            Return = 0xd,
            PageUp = 0x21,
            PageDown = 0x22,
            Home = 0x24,
            Pause = 0x91,
            Escape = 0x1b,
            Space = 0x20,
            
            Comma = 0xbc,
            Plus = 0xbb,
            Period = 0xbe,
            QuestionMark = 0xbf,
            Zero = 0x30,
            One = 0x31,
            Two = 0x32,
            Three = 0x33,
            Four = 0x34,
            Five = 0x35,
            Six = 0x36,
            Seven = 0x37,
            Eight = 0x38,
            Nine = 0x39,
            Minus = 0xbd,
            Semicolon = 0xba,
            
            Keypad0 = 0x60,
            Keypad2 = 0x62,
            Keypad4 = 0x64,
            Keypad5 = 0x65,
            Keypad6 = 0x66,
            DoubleQuote = 0xde,
            
            A = 0x41,
            B = 0x42,
            C = 0x43,
            D = 0x44,
            E = 0x45,
            F = 0x46,
            G = 0x47,
            H = 0x48,
            I = 0x49,
            J = 0x4a,
            K = 0x4b,
            L = 0x4c,
            M = 0x4d,
            N = 0x4e,
            O = 0x4f,
            P = 0x50,
            Q = 0x51,
            R = 0x52,
            S = 0x53,
            T = 0x54,
            U = 0x55,
            V = 0x56,
            W = 0x57,
            X = 0x58,
            Y = 0x59,
            Z = 0x5a,
            
            Delete = 0x8,

            // Keys that map to front panel data switches
            // Only 12 for now, for Spacewar work.
            F1 = 0x70,
            F2 = 0x71,
            F3 = 0x72,
            F4 = 0x73,
            F5 = 0x74,
            F6 = 0x75,
            F7 = 0x76,
            F8 = 0x77,
            F9 = 0x78,
            F10 = 0x79,
            F11 = 0x7a,
            F12 = 0x7b,

            // hack to toggle fullscreen.
            Insert = 0x2d,
        }        
    }
    //
    // Provides a console using SDL.
    //
    public class SDLConsole : IImlacConsole
    {
        public SDLConsole(uint scaleFactor)
        {           
            _displayList = new LinkedList<Vector>();
            _displayHash = new Dictionary<UInt64, Vector>();
            _lock = new ReaderWriterLockSlim();
            _swapLock = new ReaderWriterLockSlim();

            _vectorCache = new VectorCache(40 * 100000);    // 100,000 vectors per frame * 40 frames/sec (for persistance).  100K is wayyyyy overkill.                     
         
            _frame = 0;

            _xResolution = 1024 * scaleFactor;
            _yResolution = 1024 * scaleFactor;

            _frameTimer = new FrameTimer(40);
            _timer = new HighResTimer();

            _fullScreen = false;            

            InvokeDisplayThread();
        }        
        
        public bool IsKeyPressed
        {
            get { return _keyboardFilter.KeyLatched; }
        }

        public ImlacKey Key
        {
            get { return _keyboardFilter.Key; }
        }        

        public ImlacKeyModifiers KeyModifiers
        {
            get { return _keyboardFilter.Modifiers; }
        }

        public void UnlatchKey()
        {
            _keyboardFilter.KeyLatched = false;
        }

        public ushort DataSwitches
        {
            get { return _keyboardFilter.DataSwitches; }
        } 

        public void ClearDisplay()
        {
            _lock.EnterWriteLock();
            _displayList.Clear();
            _displayHash.Clear();
            _lock.ExitWriteLock();
            RenderCurrent();
        }

        public void MoveAbsolute(uint x, uint y, DrawingMode mode)
        {            
            //
            // Take coordinates as an 11-bit quantity (0-2048) even though we may not be displaying the full resolution.
            //             
            if (mode != DrawingMode.Off)
            {
                AddNewVector(mode, _x, _y, x, y);                
            }

            _x = x;
            _y = y;
        }

        public void MoveRelative(int dx, int dy)
        {
            //
            // If we wrap around the screen, we want to draw the correct vector

            //
        }

        public void DrawPoint(uint x, uint y)
        {
            _x = x;
            _y = y;

            AddNewVector(DrawingMode.Point, x, y, x, y);
        }

        public void FrameDone()
        {            
            RenderCurrent();
            //
            // Sync to 40hz framerate
            //
            _frameTimer.WaitForFrame();
        }        

        private void InvokeDisplayThread()
        {
            _displayThread = new System.Threading.Thread(new System.Threading.ThreadStart(DisplayThread));
            _displayThread.Start();
           
            _initEvent = new ManualResetEvent(false);

            WaitHandle[] handles = { _initEvent };

            WaitHandle.WaitAll(handles);

            Thread.Sleep(500);           

        }

        private void DisplayThread()
        {
            Video.SetVideoMode((int)_xResolution, (int)_yResolution, 32, false, false, _fullScreen, true);
            Video.WindowCaption = "Imlac PDS-1";
            _displaySurface = Video.Screen.CreateCompatibleSurface((int)_xResolution, (int)_yResolution, true);
            _displayBox = new SdlDotNet.Graphics.Primitives.Box(0, 0, (short)Video.Screen.Rectangle.Width, (short)Video.Screen.Rectangle.Height);
            
            _initEvent.Set();

            _keyboardFilter = new KeyboardFilter();
            Application.AddMessageFilter(_keyboardFilter);

            _keyboardFilter.FullScreenToggle += new EventHandler(OnFullScreenToggle);
          
            Application.Run();
        }

        void OnFullScreenToggle(object sender, EventArgs e)
        {
            _fullScreen = !_fullScreen;

            _swapLock.EnterWriteLock();
            Video.SetVideoMode((int)_xResolution, (int)_yResolution, 32, false, false, _fullScreen, true);
            _displaySurface = Video.Screen.CreateCompatibleSurface((int)_xResolution, (int)_yResolution, true);
            _swapLock.ExitWriteLock();
        }                    
       
        public void RenderCurrent()
        {            
            // Draw the current set of vectors           
            _lock.EnterReadLock();
            _frame++;

            if (_frame == 60)
            {
                double currentTime = _timer.GetCurrentTime();
                double fps = _frame / ((currentTime - _lastTime));
                _lastTime = currentTime;
                _frame = 0;

                Video.WindowCaption = String.Format("Imlac PDS-1 fps {0}", fps);
            }

            // fade out the last frame by drawing an alpha-blended black rectangle over the display.
            // (slow persistence phosphor simulation!)
            _displayBox.Draw(_displaySurface, Color.FromArgb(4, Color.Black), false, true);

            // And draw in this frame's vectors
            foreach (Vector v in _displayList)
            {
                v.Draw(_displaySurface);      
            }

            _lock.ExitReadLock();

            _swapLock.EnterReadLock();
            Video.Screen.Blit(_displaySurface);
            _displayList.Clear();                    

            Video.Screen.Update();
            _swapLock.ExitReadLock();

            
        }

        private void AddNewVector(DrawingMode mode, uint startX, uint startY, uint endX, uint endY)
        {
            // hack:  we need to make scaling work in a general fashion.  scale 11-bits down to 10 for display on 1024x1024 display
            startX = startX >> 1;
            startY = startY >> 1;

            endX = endX >> 1;
            endY = endY >> 1;

            _lock.EnterWriteLock();

            Vector newVector = _vectorCache.GetVector(mode, (short)startX, (short)(_yResolution - startY), (short)endX, (short)(_yResolution - endY));
            _displayList.AddLast(newVector);
         
            _lock.ExitWriteLock();
        }      

        private class Vector
        {
            public Vector(DrawingMode mode, uint startX, uint startY, uint endX, uint endY)
            {
                _mode = mode;                
                _line = new SdlDotNet.Graphics.Primitives.Line((short)startX, (short)startY, (short)endX, (short)endY);

                UpdateColor();
                
            }

            public void Modify(DrawingMode mode, short startX, short startY, short endX, short endY)
            {
                if (_mode != mode)
                {
                    _mode = mode;
                    UpdateColor();
                }

                _line.XPosition1 = startX;
                _line.XPosition2 = endX;
                _line.YPosition1 = startY;
                _line.YPosition2 = endY;
            }

            public void Draw(Surface displaySurface)
            {
                _line.Draw(displaySurface, _color);
            }                       

            private void UpdateColor()
            {
                switch (_mode)
                {
                    case DrawingMode.Dotted:
                    case DrawingMode.Normal:
                        _color = NormalColor;
                        break;

                    case DrawingMode.Point:
                        _color = PointColor;
                        break;

                    case DrawingMode.SGR1:
                        _color = SGRColor;
                        break;
                }
            }

            private DrawingMode _mode;
            private SdlDotNet.Graphics.Primitives.Line _line;
            private Color _color;          

            private static Color NormalColor = Color.FromArgb(196, Color.ForestGreen);
            private static Color PointColor = Color.FromArgb(255, Color.ForestGreen);
            private static Color SGRColor = Color.FromArgb(128, Color.ForestGreen);
        }

        /// <summary>
        /// To avoid continually creating/destroying Vector objects during display operations
        /// we store a "cache" of initialized objects that we continually re-use.
        /// Note that if there are more vectors onscreen than the cache holds that bad things will happen.
        /// </summary>
        private class VectorCache
        {
            public VectorCache(uint size)
            {
                Initialize(size);
            }

            public Vector GetVector(DrawingMode mode, short startX, short startY, short endX, short endY)
            {
                Vector next = _cache[_nextVector % _cache.Length];
                next.Modify(mode, startX, startY, endX, endY);                

                _nextVector++;

                return next;
            }

            private void Initialize(uint size)
            {
                _cache = new Vector[size];

                for (uint i = 0; i < size; i++)
                {
                    _cache[i] = new Vector(DrawingMode.Hidden, 0, 0, 0, 0);
                }

                _nextVector = 0;
            }

            private Vector[] _cache;
            private uint _nextVector;
        }

        private System.Threading.Thread _displayThread;
        private Surface _displaySurface;
        private SdlDotNet.Graphics.Primitives.Box _displayBox;

        private ManualResetEvent _initEvent;
        
        private uint _x;
        private uint _y;

        private uint _xResolution;
        private uint _yResolution;

        private bool _fullScreen;

        private LinkedList<Vector> _displayList;
        private Dictionary<UInt64, Vector> _displayHash;
        private VectorCache _vectorCache;
        private System.Threading.ReaderWriterLockSlim _lock;
        private System.Threading.ReaderWriterLockSlim _swapLock;

        private const int _initBrightness = 196;

        // keyboard input data
        private KeyboardFilter _keyboardFilter;

        private uint _frame;
        private double _lastTime;

        // Framerate manaement
        FrameTimer _frameTimer;
        HighResTimer _timer;
        
    }
}
