﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;

using imlac.IO;

namespace imlac
{
    //
    // Provides a console using only Winforms.  This is a fallback for when other methods are not available.
    // It is not fast.
    //
    public class WinformsConsole : IImlacConsole
    {
        public WinformsConsole()
        {
            _vectorPens = new List<Pen>();
            
            for (int i = 0; i < 16; i++)
            {
                _vectorPens.Add(new Pen(Color.FromArgb(i * 8, Color.LightGreen)));
            }            

            _displayList = new LinkedList<Vector>();
            _lock = new ReaderWriterLockSlim();

            InvokeDisplayThread();            
        }
         
        static WinformsConsole()
        {
            BuildKeyMappings();
        }

        public bool IsKeyPressed
        {
            get 
            {                
                return _keyLatched;            
            }
        }

        public ImlacKey Key
        {
            get { return _pressedKey; }
        }        

        public ImlacKeyModifiers KeyModifiers
        {
            get { return _keyModifiers; }
        }

        public void UnlatchKey()
        {
            _keyLatched = false;
        }

        public ushort DataSwitches
        {
            get { return 0; }
        }            

        public void ClearDisplay()
        {
            _lock.EnterWriteLock();
            _displayList.Clear();
            _lock.ExitWriteLock();
            RenderCurrent();
        }

        public void MoveAbsolute(uint x, uint y, DrawingMode mode)
        {
            //
            // Scale coordinates by half (2048 pt -> 1024 pt)
            // 
            x = x >> 1;
            y = y >> 1;

            if (mode != DrawingMode.Off)
            {
                _lock.EnterWriteLock();
                _displayList.AddLast(new Vector(mode, _x, 1024 - _y, x, 1024 - y));
                _lock.ExitWriteLock();
            }

            _x = x;
            _y = y;
        }

        public void MoveRelative(int dx, int dy)
        {
            //
            // If we wrap around the screen, we want to draw the correct vector

            //
        }

        public void DrawPoint(uint x, uint y)
        {
            //
            // Scale coordinates by half (2048 pt -> 1024 pt)
            // 
            x = x >> 1;
            y = y >> 1;

            _lock.EnterWriteLock();
            _displayList.AddLast(new Vector(DrawingMode.Normal, _x, 1024 - _y, _x, 1024 - _y));   // this is ugly
            _lock.ExitWriteLock();
        }

        public void FrameDone()
        {
            ExpireDisplayList();
            RenderCurrent();            
        }

        private void InvokeDisplayThread()
        {
            _displayThread = new System.Threading.Thread(new System.Threading.ThreadStart(DisplayThread));
            _displayThread.Start();

            _initEvent = new ManualResetEvent(false);

            WaitHandle[] handles = { _initEvent };

            WaitHandle.WaitAll(handles);

            Thread.Sleep(500);
        }

        private void DisplayThread()
        {
            _display = new Form();
            _display.CreateControl();
            _display.BackColor = Color.Black;
            _display.Text = "Imlac PDS-1";
            _display.ControlBox = false;
            _display.ClientSize = new Size(1024, 1024);
            _display.SizeGripStyle = SizeGripStyle.Hide;
            _display.WindowState = FormWindowState.Normal;
            _display.KeyPreview = true;            
            _display.KeyUp += new KeyEventHandler(OnKeyUp);
            _display.KeyDown += new KeyEventHandler(OnKeyDown);        
            
            //_display.MouseWheel += new MouseEventHandler(OnMouseWheel);
            

            _buffer = new Bitmap(1024, 1024, PixelFormat.Format32bppArgb);            

            _dispBox = new PictureBox();            
            _dispBox.Size = new Size(1024, 1024);
            _dispBox.Cursor = Cursors.Cross;
            _dispBox.Paint += new PaintEventHandler(OnPaint);
            _dispBox.Enabled = false;
            
            //_dispBox.MouseDown += new MouseEventHandler(OnMouseDown);
            //_dispBox.MouseUp += new MouseEventHandler(OnMouseUp);
            //_dispBox.MouseMove += new MouseEventHandler(OnMouseMove);

            _display.Controls.Add(_dispBox);

            BufferedGraphicsContext context = BufferedGraphicsManager.Current;
            context.MaximumBuffer = new Size(1024, 1024);

            _graphics = context.Allocate(_dispBox.CreateGraphics(), new Rectangle(0, 0, 1024, 1024));

            _initEvent.Set();

            _display.ShowDialog();     

        }        

        void OnKeyDown(object sender, KeyEventArgs e)
        {
            //e.SuppressKeyPress = true;
            //e.Handled = true;
            if (_keyMappings.ContainsKey(e.KeyCode))
            {
                _pressedKey = _keyMappings[e.KeyCode];
            }
            else
            {
                _pressedKey = ImlacKey.Invalid;
            }

            _keyModifiers = ImlacKeyModifiers.None;

            if ((e.Modifiers & Keys.Control) != 0)
            {
                _keyModifiers |= ImlacKeyModifiers.Ctrl;
            }

            if ((e.Modifiers & Keys.Shift) != 0)
            {
                _keyModifiers |= ImlacKeyModifiers.Shift;
            }

            if ((e.Modifiers & Keys.Alt) != 0)
            {
                _keyModifiers |= ImlacKeyModifiers.Rept;
            }

            _keyLatched = (_pressedKey != ImlacKey.Invalid);                                       
        }

        void OnKeyUp(object sender, KeyEventArgs e)
        {
            //e.SuppressKeyPress = true;
            //e.Handled = true;            
        }
       
        public void RenderCurrent()        // public for debugging purposes
        {
            // Draw the current set of vectors
            Graphics g = _graphics.Graphics;
            g.FillRectangle(Brushes.Black, 0, 0, 1024, 1024);

            _lock.EnterReadLock();
            foreach (Vector v in _displayList)
            {
                if (v.Mode != DrawingMode.Off)
                {
                    Pen pen = _vectorPens[v.TTL * 2];

                    pen.DashStyle = (v.Mode == DrawingMode.Normal || v.Mode == DrawingMode.SGR1) ? System.Drawing.Drawing2D.DashStyle.Solid : System.Drawing.Drawing2D.DashStyle.Dot;

                    if (v.StartX != v.EndX || v.StartY != v.EndY)
                    {
                        g.DrawLine(pen, v.StartX, v.StartY, v.EndX, v.EndY);
                    }
                    else
                    {
                        // TODO: remove this hack
                        pen = _vectorPens[v.TTL * 3];
                        g.DrawLine(pen, v.StartX, v.StartY, v.EndX + 1, v.EndY);
                    }
                }
            }
            _lock.ExitReadLock();

            _graphics.Render(Graphics.FromHwnd(_dispBox.Handle));
        }

        private void OnPaint(object sender, PaintEventArgs e)
        {
            //_graphics.Render(e.Graphics);
        }

        private void ExpireDisplayList()
        {
            // We have an ordered list from oldest to newest, we decrement the TTL counter and when it reaches
            // zero we remove it from the list.
            _lock.EnterWriteLock();
            LinkedListNode<Vector> current = _displayList.First;

            while (current != null)
            {
                Vector v = current.Value;                
                v.TTL--;

                current = current.Next;

                if (v.TTL <= 0)
                {
                    v.TTL = 0;
                   _displayList.RemoveFirst();
                }                
            }
           
            _lock.ExitWriteLock();
        }

        private class Vector
        {
            public Vector(DrawingMode mode, uint startX, uint startY, uint endX, uint endY)
            {
                Mode = mode;
                StartX = startX;
                StartY = startY;
                EndX = endX;
                EndY = endY;
                TTL = mode == DrawingMode.SGR1 ? 1 : 5;
            }

            public DrawingMode Mode;
            public uint StartX;
            public uint StartY;
            public uint EndX;
            public uint EndY;
            public int TTL;
        }

        private System.Threading.Thread _displayThread;
        private Form _display;
        private Bitmap _buffer;
        private PictureBox _dispBox;
        private BufferedGraphics _graphics;

        private ManualResetEvent _initEvent;
        
        private List<Pen> _vectorPens;
        private uint _x;
        private uint _y;

        private LinkedList<Vector> _displayList;
        private System.Threading.ReaderWriterLockSlim _lock;

        // keyboard input data
        bool _keyLatched;
        ImlacKey _pressedKey;
        ImlacKeyModifiers _keyModifiers;

        private static void BuildKeyMappings()
        {
            _keyMappings = new Dictionary<Keys, ImlacKey>();

            _keyMappings.Add(Keys.End, ImlacKey.DataXmit);
            _keyMappings.Add(Keys.Down, ImlacKey.Down);
            _keyMappings.Add(Keys.Right, ImlacKey.Right);
            _keyMappings.Add(Keys.Up, ImlacKey.Up);
            _keyMappings.Add(Keys.Left, ImlacKey.Left);
            _keyMappings.Add(Keys.Tab, ImlacKey.Tab);            
            _keyMappings.Add(Keys.Return, ImlacKey.CR);
            _keyMappings.Add(Keys.PageUp, ImlacKey.FF);
            _keyMappings.Add(Keys.PageDown, ImlacKey.PageXmit);
            _keyMappings.Add(Keys.Home, ImlacKey.Home);
            _keyMappings.Add(Keys.Pause, ImlacKey.Brk);
            _keyMappings.Add(Keys.Escape, ImlacKey.Esc);
            _keyMappings.Add(Keys.Space, ImlacKey.Space);
                
            _keyMappings.Add(Keys.Oemcomma, ImlacKey.Comma);
            _keyMappings.Add(Keys.Oemplus, ImlacKey.Minus);
            _keyMappings.Add(Keys.OemPeriod, ImlacKey.Period);
            _keyMappings.Add(Keys.OemQuestion, ImlacKey.Slash);
            _keyMappings.Add(Keys.D0, ImlacKey.K0);
            _keyMappings.Add(Keys.D1, ImlacKey.K1);
            _keyMappings.Add(Keys.D2, ImlacKey.K2);
            _keyMappings.Add(Keys.D3, ImlacKey.K3);
            _keyMappings.Add(Keys.D4, ImlacKey.K4);
            _keyMappings.Add(Keys.D5, ImlacKey.K5);
            _keyMappings.Add(Keys.D6, ImlacKey.K6);
            _keyMappings.Add(Keys.D7, ImlacKey.K7);
            _keyMappings.Add(Keys.D8, ImlacKey.K8);
            _keyMappings.Add(Keys.D9, ImlacKey.K9);            
            _keyMappings.Add(Keys.OemMinus, ImlacKey.Colon);
            _keyMappings.Add(Keys.OemSemicolon, ImlacKey.Semicolon);

            _keyMappings.Add(Keys.NumPad0, ImlacKey.D0);
            _keyMappings.Add(Keys.NumPad2, ImlacKey.D2);
            _keyMappings.Add(Keys.NumPad4, ImlacKey.D4);
            _keyMappings.Add(Keys.NumPad5, ImlacKey.D5);
            _keyMappings.Add(Keys.NumPad6, ImlacKey.D6);
            _keyMappings.Add(Keys.OemQuotes, ImlacKey.Unlabeled);

            _keyMappings.Add(Keys.A, ImlacKey.A);
            _keyMappings.Add(Keys.B, ImlacKey.B);
            _keyMappings.Add(Keys.C, ImlacKey.C);
            _keyMappings.Add(Keys.D, ImlacKey.D);
            _keyMappings.Add(Keys.E, ImlacKey.E);
            _keyMappings.Add(Keys.F, ImlacKey.F);
            _keyMappings.Add(Keys.G, ImlacKey.G);
            _keyMappings.Add(Keys.H, ImlacKey.H);
            _keyMappings.Add(Keys.I, ImlacKey.I);
            _keyMappings.Add(Keys.J, ImlacKey.J);
            _keyMappings.Add(Keys.K, ImlacKey.K);
            _keyMappings.Add(Keys.L, ImlacKey.L);
            _keyMappings.Add(Keys.M, ImlacKey.M);
            _keyMappings.Add(Keys.N, ImlacKey.N);
            _keyMappings.Add(Keys.O, ImlacKey.O);
            _keyMappings.Add(Keys.P, ImlacKey.P);
            _keyMappings.Add(Keys.Q, ImlacKey.Q);
            _keyMappings.Add(Keys.R, ImlacKey.R);
            _keyMappings.Add(Keys.S, ImlacKey.S);
            _keyMappings.Add(Keys.T, ImlacKey.T);
            _keyMappings.Add(Keys.U, ImlacKey.U);
            _keyMappings.Add(Keys.V, ImlacKey.V);
            _keyMappings.Add(Keys.W, ImlacKey.W);
            _keyMappings.Add(Keys.X, ImlacKey.X);
            _keyMappings.Add(Keys.Y, ImlacKey.Y);
            _keyMappings.Add(Keys.Z, ImlacKey.Z);
            

            _keyMappings.Add(Keys.Delete, ImlacKey.Del);
        }

        private static Dictionary<Keys, ImlacKey> _keyMappings;
    }
}
